/*
	Copyright (C) 2004 Jonas Lindholm

	This software was developed by Jonas Lindholm, jlhm@usa.net

	History

	V1.0		Jonas Lindholm	2004-05-14

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>

#include <ldap.h>

#include <starlet.h>
#include <ssdef.h>
#include <stsdef.h>
#include <psldef.h>
#include <jpidef.h>
#include <descrip.h>

#include <pthread.h>

#include "lgildap.h"

#include "server.h"


/*
	Description:

	This routine try to establish a session with the LDAP server
	using the identity and credential supplied by the user.

	If the job type is network and the password check fail the assword
	is converted to lower case and the check is done again but only if
	the logical name LGI_LDAP_PWD_NETWORK_IGNORE_CASe is defined.

	INPUT:		session handler
			LGI message

	OUTPUT:		SS$_NORMAL if the credential was ok
			LLSRV_INVCRED if the credential was incorrect

	IMPLICIT:	The identity must be valid, flag identityok must be set
*/

int thread_cmd_authenticate(LGI_SESS_T *sess, LIM_ENTRY_T *le) {
	int		stat;

        le->lim.rplyid = le->lim.reqid;
        le->lim.reqid++;
	le->lim.status = LLSRV_INVCRED;

	if (sess->dn != NULL && sess->flags.identityok) {

		pthread_mutex_lock(&ldapmtx);

		stat = ldap_simple_bind_s(sess->ldap, sess->dn, le->lim.request.authenticate.credential);

		if (stat == 0) {

			pthread_mutex_unlock(&ldapmtx);

			synch_password(sess, le->lim.request.authenticate.credential);

			le->lim.status = SS$_NORMAL;

			sess->flags.credentialok = 1;

			return SS$_NORMAL;

		}

		/*
			We have a problem with network login and the case of the password.
			The encrypted password stored in LDAP is the result from the password
			specified with lowercase and uppercase characters.
			However, for network logins the password is always in uppercase.
			Should the password has been specified with something else than just
			uppercase characters the login test will fail.
			One quick fix for network logins is to change it to lowercase in case
			the uppercase password check failed and try the login check again.
			This will however not work when the user has a password with a mix of
			uppercase and lowercase characters.
		*/

		if (sess->jobtype == JPI$K_NETWORK) { /* test with lowercase password */
			char			d[256];

			if ($VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_PWD_NETWORK_IGNORE_CASE", d, sizeof(d) - 1, PSL$C_EXEC))) {
				char		*p = le->lim.request.authenticate.credential;

				for(; *p != 0; *p = tolower(*p), p++);

	                	stat = ldap_simple_bind_s(sess->ldap, sess->dn, le->lim.request.authenticate.credential);

				if (stat == 0) {

					pthread_mutex_unlock(&ldapmtx);

					synch_password(sess, le->lim.request.authenticate.credential);

					le->lim.status = SS$_NORMAL;

					sess->flags.credentialok = 1;

					return SS$_NORMAL;

				}
			}
		}

		pthread_mutex_unlock(&ldapmtx);
        }

	return SS$_ABORT;
}
