/*
	Copyright (C) 2004 Jonas Lindholm

	This software was developed by Jonas Lindholm, jlhm@usa.net

	History

	V1.0		Jonas Lindholm	2004-05-14

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	ldap.h include all standard headers such as string.h
*/
#include <string.h>

#include <ldap.h>

#include <starlet.h>
#include <ssdef.h>
#include <stsdef.h>
#include <psldef.h>
#include <lnmdef.h>
#include <jpidef.h>
#include <uaidef.h>
#include <descrip.h>

#include <lib$routines.h>

#include <rms.h>
#include <fab.h>
#include <rab.h>
#include <fabdef.h>
#include <rabdef.h>
#include <prcdef.h>

#include <accdef.h>
#include <dvidef.h>
#include <cmbdef.h>
#include <iodef.h>
#include <nsadef.h>


#include <pthread.h>

#include "lgi.h"
#include "lgildap.h"

#include "server.h"

#include "uafdef.h"


typedef struct uafact_s {
	int unsigned		qhdr[4];
	int			stat;
	char			username[32 + 1];
	char			password[32 + 1];
	LGI_SESS_T		*sess;
} UAFACT_T;                                                                    


typedef struct uafpriv_s {
	struct {
		short		minver;
		short		majver;
	} version;

	int unsigned		magic;   /* magic value for consistency check */
	char			dn[128]; /* LDAP DN used when creating the user */
	char			uid[32]; /* UID used in LDAP for this user */
} UAFPRIV_T;

#define			UAFPRIV_MIN	0
#define			UAFPRIV_MAJ	1
#define			UAFPRIV_MAGIC	0x4c444150  /* LDAP as HEX string */

int unsigned		addqh[4] = { 0, 0, 0, 0 };
int unsigned		verqh[4] = { 0, 0, 0, 0 };
int unsigned		pwdqh[4] = { 0, 0, 0, 0 };

int			add_user_ok = 0;

pthread_t		thrmod;
pthread_t		thradd;
pthread_t		thrrem;

pthread_cond_t		cndw;
pthread_mutex_t		mtxw;
pthread_cond_t		cndaddw;
pthread_mutex_t		mtxaddw;
pthread_mutex_t		mtxuai;


/*
	Description:

	This routine create an audit and alarm event.
	It is invoked when a user is added, deleted or modified by this program.
	The DN is logged together with the system service name, LGI LDAP Server.

	This make it possible to keep track of all actions done with this process
	in the SYSUAF.DAT file using RMS routines.

	INPUT:		event subtype
			UAF record
			current DN
			new DN

	OUTPUT:		status of sys$audit
			SS$_ABORT of subtype is invalid
*/

int do_audit(int subtype, char *username, char *curdn, char*newdn) {
	int			stat;
	int			evtype = NSA$C_MSG_SYSUAF;
	char			auditname[] = "SECURITY";
	char			ssn[] = "LGI LDAP Server";
	char			fn[] = "Distinguished Name";
	ITEM3			itmlst[] = { { sizeof(evtype), NSA$_EVENT_TYPE, &evtype, 0 },
					     { sizeof(subtype), NSA$_EVENT_SUBTYPE, &subtype, 0 },
					     { strlen(auditname), NSA$_AUDIT_NAME, &auditname, 0 },
					     { strlen(auditname), NSA$_ALARM_NAME, &auditname, 0 },
					     { strlen(ssn), NSA$_SYSTEM_SERVICE_NAME, &ssn, 0 },
					     { 4, NSA$_CHAIN, 0, 0 } };

	switch (subtype) {
		case NSA$C_SYSUAF_ADD: {
			ITEM3		itmlst2[] = { { strlen(username), NSA$_UAF_ADD, username, 0 },
						      { strlen(fn), NSA$_FIELD_NAME, fn, 0 },
						      { strlen(newdn), NSA$_NEW_DATA, newdn, 0 },
						      { 0, 0, 0, 0 } };
			itmlst[5].bufadr = &itmlst2;
			stat = sys$audit_eventw(0, 0, &itmlst, 0, 0, 0);

			break;
		}

		case NSA$C_SYSUAF_DELETE: {
			ITEM3		itmlst2[] = { { strlen(username), NSA$_UAF_ADD, username, 0 },
						      { strlen(fn), NSA$_FIELD_NAME, fn, 0 },
						      { strlen(curdn), NSA$_ORIGINAL_DATA, curdn, 0 },
						      { 0, 0, 0, 0 } };
			itmlst[5].bufadr = &itmlst2;
			stat = sys$audit_eventw(0, 0, &itmlst, 0, 0, 0);

		      	break;
		}

		case NSA$C_SYSUAF_MODIFY: {
			ITEM3		itmlst2[] = { { strlen(username), NSA$_UAF_MODIFY, username, 0 },
						      { strlen(fn), NSA$_FIELD_NAME, fn, 0 },
						      { strlen(curdn), NSA$_ORIGINAL_DATA, curdn, 0 },
						      { strlen(newdn), NSA$_NEW_DATA, newdn, 0 },
						      { 0, 0, 0, 0 } };
			itmlst[5].bufadr = &itmlst2;
			stat = sys$audit_eventw(0, 0, &itmlst, 0, 0, 0);

			break;
		}
		default: {
			return SS$_ABORT;
		}
	}

	return stat;
}


/*
	Description:

	This routine is used to open and connect FAB and RAB structures to SYSUAF.DAT.
	It use only executive defined logical names for SYSUAF.DAT.

	INPUT:		FAB
			RAB

	OUTPUT:		RMS$_NORMAL if SYSUAF.DAT open and RAB connected
			any RMS error
*/

int open_sysuaf(struct FAB *fab , struct RAB *rab) {
	int		stat;
	char		sysuafdef[] = "SYS$SYSTEM:.DAT";
	char	       	sysuaf[] = "SYSUAF";

	fab->fab$v_lnm_mode = PSL$C_EXEC;
	fab->fab$l_dna = (char *)&sysuafdef;
	fab->fab$b_dns = strlen(sysuafdef);
	fab->fab$l_fna = (char *)&sysuaf;
	fab->fab$b_fns = strlen(sysuaf);
        fab->fab$b_fac = FAB$M_DEL | FAB$M_GET | FAB$M_PUT | FAB$M_UPD;
        fab->fab$b_shr = FAB$M_SHRDEL | FAB$M_SHRGET | FAB$M_SHRPUT | FAB$M_SHRUPD;

	if ($VMS_STATUS_SUCCESS(stat = sys$open(fab, 0, 0))) {

		rab->rab$l_fab = fab;

		if (!$VMS_STATUS_SUCCESS(stat = sys$connect(rab , 0, 0))) {

			sys$close(fab, 0, 0);

		}
	}

	return stat;
}


/*
	Description:

	This routine disconnect and close SYSUAF.DAT

	INPUT:		FAB
			RAB

	OUTPUT:		none
*/

void close_sysuaf(struct FAB *fab, struct RAB *rab) {

	sys$disconnect(rab , 0, 0);
	sys$close(fab, 0, 0);

}


/*
	Description:

	This routine fetch the owner (cn), default device (homeDirectory), default directory (homeDirectory)
	and UIC (uid & gid).

	INPUT:		pointer to ldap structure
			DN
			uid to fetch
			owner, must be 32 bytes or longer
			defdev, must be 32 bytes or longer
			defdir, must be 64 bytes or longer
			UIC

	OUTPUT:		SS$_NORMAL, updated owner, defdev, defdir and UIC, if everything is ok
			ldap set to NULL is any LDAP call fail
			LLSRV_LDAPUSERMISS is DN/uid is missing
*/

get_ldap_user_data(LDAP **ldap, char *dn, char *uid, char *owner, char *defdev, char *defdir, short unsigned uic[]) {
	int			stat;
	int			retstat = LLSRV_LDAPUSERMISS;
	int			msgid;
	char			filter[64] = "";
	char			*attrs[] = { "cn", "uidNumber", "gidNumber", "homeDirectory", NULL };
	LDAPMessage		*res;

	if (*ldap == NULL) { return retstat; }

	pthread_mutex_lock(&ldapmtx);

	sprintf(filter, "(uid=%s)", uid);

	if ((msgid = ldap_search(*ldap, dn, LDAP_SCOPE_BASE, filter, attrs, 0)) > 0) {
		struct timeval          to = { 0, 0 };

		while ((stat = ldap_result(*ldap, msgid, 1, &to, &res)) <= 0) {
			struct timespec         wait = { 0, 50000 };

			pthread_mutex_unlock(&ldapmtx);
			pthread_delay_np(&wait);
			pthread_mutex_lock(&ldapmtx);

		}

		if (stat > 0) {
			LDAPMessage             *e = (void *)1; /* in case something goes wrong with ldap_first_entry... */

			if ((e = ldap_first_entry(*ldap, res)) != NULL) {
				int		okcnt = 0; /* must reach 4 to be ok, all required attributes ok */
				char 		**attr;

				if ((attr = ldap_get_values(*ldap, e, "cn")) != NULL) {

					strncpy(owner, attr[0], 32);
					ldap_value_free(attr);
					okcnt++;

				}

				if ((attr = ldap_get_values(*ldap, e, "uidNumber")) != NULL) {
					char		offset[7] = "010000";
					int		offs = 010000;

					do_trnlnm("LGI_LDAP_ADD_USER_UID_OFFSET", offset, sizeof(offset) - 1, PSL$C_EXEC);
					stat = sscanf(offset, "%i", &offs);
					if (stat == 0) { lib$signal(LLSRV_INVVALUE, 1, "LGI_LDAP_ADD_USER_UID_OFFSET"); }
					uic[0] = atoi(attr[0]) + offs;
					ldap_value_free(attr);
					okcnt++;

				}

				if ((attr = ldap_get_values(*ldap, e, "gidNumber")) != NULL) {
					char		offset[7] = "01000";
					int		offs = 01000;

					do_trnlnm("LGI_LDAP_ADD_USER_GID_OFFSET", offset, sizeof(offset) - 1, PSL$C_EXEC);
					stat = sscanf(offset, "%i", &offs);
					if (stat == 0) { lib$signal(LLSRV_INVVALUE, 1, "LGI_LDAP_ADD_USER_GID_OFFSET"); }
					uic[1] = atoi(attr[0]) + offs;
					ldap_value_free(attr);
					okcnt++;

				}

				if ((attr = ldap_get_values(*ldap, e, "homeDirectory")) != NULL) {

					translate_unix2vms(attr[0], defdev, 31, defdir, 63);
					do_trnlnm(defdev, defdev, 31, PSL$C_EXEC);
					ldap_value_free(attr);
					okcnt++;

				}

				ldap_msgfree(e);

				if (okcnt == 4) { retstat = SS$_NORMAL; } /* only return if all attributes were fetched */

			} else {

				ldap_unbind(*ldap);
				*ldap = NULL;

			}

			ldap_msgfree(res);

		} else {

			ldap_unbind(*ldap);
			*ldap = NULL;

		}
	}

	pthread_mutex_unlock(&ldapmtx);

	return retstat;
}


/*
	Description:

	This routine is used to verify the existence of a user in SYSUAF.DAT.
	If the user exists and has a uafpriv structure in the user data area
	a check is done to see if the used DN for the session is the same DN
	as was used to create the user.

	If the DN has changed the rutine will fetch the cn, uidNumber, gidNumber
	and homeDirectory and update the user. An audit alarm is also generated.

	INPUT:		session handler
			identity (username)

	OUTPUT:		SS$_NORMAL if the user exists in SYSUAF.DAT
			error when user doesn't exist or same other error.
*/

int uaf_verify(LGI_SESS_T *sess, char *identity) {
	int			stat;
	short			udlen;
	short unsigned		curuic[2];
	IOSB			iosb;
	UAFPRIV_T		uafpriv;
	ITEM3			itmlst[] = { {sizeof(uafpriv), UAI$_USER_DATA, &uafpriv, &udlen },
					     { sizeof(curuic), UAI$_UIC, &curuic, 0 },
					     { 0, 0, 0, 0 } };

	$DESCRIPTOR(usrnamdsc, identity);

	usrnamdsc.dsc$w_length = strlen(identity);

	stat = sys$getuai(0, 0, &usrnamdsc, &itmlst, 0, 0, 0);

	if ($VMS_STATUS_SUCCESS(stat) && (udlen >= sizeof(uafpriv)) && (uafpriv.magic = UAFPRIV_MAGIC)) {

		if (strcmp(sess->dn, uafpriv.dn) != 0) { /* DN from LDAP server is not the same as in SYSUAF.DAT */
			char			defdev[33] = ".USERS"; /* save first byte for length */
			char			defdir[65] = ".USER";
			char			owner[33] = ".Unknown LDAP user";
			short unsigned		newuic[2] = { 01000, 01000 }; /* default UIC */
			UAF			uafrec;

			stat = get_ldap_user_data(&sess->ldap, sess->dn, identity, owner + 1, defdev + 1, defdir + 1, newuic);

			if ($VMS_STATUS_SUCCESS(stat)) {
				ITEM3		itmlsts[] = { { sizeof(defdev) - 1, UAI$_DEFDEV, &defdev, 0 },
							      { sizeof(defdir) - 1, UAI$_DEFDIR, &defdir, 0 },
							      { sizeof(owner) - 1,  UAI$_OWNER,  &owner,  0 },
							      { sizeof(newuic),     UAI$_UIC,    &newuic, 0 },
							      { sizeof(uafpriv),    UAI$_USER_DATA, &uafpriv, 0 },
							      { 0, 0, 0, 0 } };

				defdev[0] = strlen(defdev + 1);
				defdir[0] = strlen(defdir + 1);
				owner[0] = strlen(owner + 1);

				do_audit(NSA$C_SYSUAF_MODIFY, identity, uafpriv.dn, sess->dn);

				strncpy(uafpriv.dn, sess->dn, sizeof(uafpriv.dn));
				strncpy(uafpriv.uid, identity, sizeof(uafpriv.uid));

				stat = sys$setuai(0, 0, &usrnamdsc, &itmlsts, 0, 0, 0);

				if (curuic[0] != newuic[0] || curuic[1] != newuic[1]) { /* modify identifier value if new UIC */

					sys$mod_ident((curuic[0] | (curuic[1] << 16)), 0, 0, 0, (newuic[0] | (newuic[1] << 16)));

				}
			}
		}
	}

	return stat;
}


/*
	Description:

	This routine is used to update the local password hash in SYSUAF.DAT
	for the user.
	The update is done if the password used was correct and the logical
	name LGI_LDAP_PASSWORD_SYNCH is defined.
	The hashed password value is updated only if the hash value for the
	correct password is something else than the currently stored value.

	INPUT:		session handler
			password

	OUTPUT:		SS$_NORMAL is everything was ok
			error from sys$getuai, sys$setuai or sys$hash_password
*/

int uaf_pwd_synch(LGI_SESS_T *sess, char *password) {
	int			stat;
	int unsigned		salt;
	int unsigned		pwdhash[2];
	int unsigned		encalg;
	char			username[sizeof(sess->identity)];
	char			*t = sess->identity;
	char			*d = username;
	IOSB			iosb;
	ITEM3			itmlst[4] = { { sizeof(pwdhash), UAI$_PWD, &pwdhash, 0}, { sizeof(salt), UAI$_SALT, &salt, 0}, 
					      { sizeof(encalg), UAI$_ENCRYPT, &encalg, 0}, { 0, 0, 0, 0 } };

	/* use username as dummy variable */

	if (!$VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_PASSWORD_SYNCH", (char *)&username, sizeof(username) - 1, PSL$C_EXEC))) {

		return SS$_NORMAL;

	}

	for(; *t != 0; *d = toupper(*t), t++, d++);
	*d = 0;

	$DESCRIPTOR(usrnamdsc, username);

	usrnamdsc.dsc$w_length = strlen(username);

	stat = sys$getuai(0, 0, &usrnamdsc, &itmlst, 0, 0, 0);

	if ($VMS_STATUS_SUCCESS(stat)) {
		int unsigned		pwdh[2];
		char			pwd[256];

		for(d = pwd, t = password; *d = toupper(*t); t++, d++);
		*d = 0;

		$DESCRIPTOR(pwddsc, pwd);

		pwddsc.dsc$w_length = strlen(pwd);
		stat = sys$hash_password(&pwddsc, encalg, salt, &usrnamdsc, &pwdh);

		if ($VMS_STATUS_SUCCESS(stat)) {

			if (pwdh[0] != pwdhash[0] || pwdh[1] != pwdhash[1]) {

				pwdhash[0] = pwdh[0];
				pwdhash[1] = pwdh[1];

				memset(&itmlst[1], 0, sizeof(itmlst[1]));

				stat = sys$setuai(0, 0, &usrnamdsc, &itmlst, 0, 0, 0);

			}
		}
	}

	return stat;
}


/*
	Description:

	This AST rutine is invoked when any command procedure times out.
	It signal a condition variable that will wake up the thread waiting
	for the command procedure to finish.

	INPUT:		condition variable

	OUTPUT		none
*/

void qio_done_ast(pthread_cond_t *cndw) {

	pthread_cond_signal_int_np(cndw);

}


/*
	Description:

	This routine creates a process and run the specified command procedure.
	It will wait for the cmmand procedure to finish in N seconds or otherwise
	the process is stopped.

	INPUT:		command procedure
			timeout in seconds
			username
			default device
			default directory
			UIC

	OUTPUT:		any error caused by sys$qio, sys$creprc or sys$crembx
			SS$_ABORT if process timeout.
*/

int do_command(char *actrtn, int timeout, char *username, char *defdev, char *defdir, short unsigned uic[]) {
	int			stat;
	int unsigned		trmunit;
	int unsigned		pid;
	int unsigned		iosb[2] = { SS$_NORMAL, 0 };
	short unsigned		outchan = 0;
	short unsigned		trmchan = 0;
	char			mbxnam[65];
	char			str[1024];
	char			logout[] = "LOGOUT_";
	struct timespec		wait = { timeout, 0 };
	struct accdef		acc;
	pthread_cond_t		cndt;
	pthread_mutex_t		mtxt;
	$DESCRIPTOR(mbxnamdsc, mbxnam);
	$DESCRIPTOR(loginoutdsc, "SYS$SYSTEM:LOGINOUT.EXE");
	$DESCRIPTOR(outdsc, "SYS$MANAGER:LGILDAPSRV_USER_ACTION.LOG");
	ITEM3			itmlst[] = { { sizeof(mbxnam) - 1, DVI$_ALLDEVNAM, &mbxnam, &mbxnamdsc.dsc$w_length },
					     { sizeof(trmunit), DVI$_UNIT, &trmunit, 0}, { 0, 0, 0, 0 } };

	pthread_cond_init(&cndt, NULL);
	pthread_mutex_init(&mtxt, NULL);

	if ($VMS_STATUS_SUCCESS(stat = sys$crembx(0, &outchan, 1024, 1024 * 5, 0, 0, 0, CMB$M_WRITEONLY, 0))) {

		if ($VMS_STATUS_SUCCESS(stat = sys$crembx(0, &trmchan, ACC$K_TERMLEN * 2, ACC$K_TERMLEN * 2, 0, 0, 0, CMB$M_READONLY, 0))) {

			if ($VMS_STATUS_SUCCESS(stat = sys$getdviw(0, trmchan, 0, &itmlst, &iosb, 0, 0, 0))) {

				memset(&itmlst[1], 0, sizeof(itmlst[1]));

				stat = sys$getdviw(0, outchan, 0, &itmlst, &iosb, 0, 0, 0);

			}
		}
	}

	if ($VMS_STATUS_SUCCESS(stat) &&
	    $VMS_STATUS_SUCCESS(stat = sys$creprc(&pid, &loginoutdsc, &mbxnamdsc, &outdsc, 0, 0, 0, 0, 0, 0, trmunit, 0, 0, 0, 0))) {
		struct timespec		abstime;

		sprintf(str, "@%s", actrtn);
        	stat = sys$qiow(0, outchan, IO$_WRITEVBLK | IO$M_NOW, &iosb, 0, 0, &str, strlen(str), 0, 0, 0, 0);

        	stat = sys$qiow(0, outchan, IO$_WRITEVBLK | IO$M_NOW, &iosb, 0, 0, username, strlen(username), 0, 0, 0, 0);

        	stat = sys$qiow(0, outchan, IO$_WRITEVBLK | IO$M_NOW, &iosb, 0, 0, defdev, strlen(defdev), 0, 0, 0, 0);

        	stat = sys$qiow(0, outchan, IO$_WRITEVBLK | IO$M_NOW, &iosb, 0, 0, defdir, strlen(defdir), 0, 0, 0, 0);

		sprintf(str, "[%o,%o]", uic[1], uic[0]);
        	stat = sys$qiow(0, outchan, IO$_WRITEVBLK | IO$M_NOW, &iosb, 0, 0, str, strlen(str), 0, 0, 0, 0);

        	stat = sys$qiow(0, outchan, IO$_WRITEVBLK | IO$M_NOW, &iosb, 0, 0, logout, strlen(logout), 0, 0, 0, 0);

		pthread_get_expiration_np(&wait, &abstime);

		stat = sys$qio(0, trmchan, IO$_READVBLK, &iosb, &qio_done_ast, &cndt, &acc, sizeof(acc), 0, 0, 0, 0);

		pthread_cond_timedwait(&cndt, &mtxt, &abstime);

		sys$delprc(&pid, 0, 0); /* in case process still running */

	}

	sys$dassgn(trmchan);
	sys$dassgn(outchan);

	if ($VMS_STATUS_SUCCESS(stat) && !$VMS_STATUS_SUCCESS(iosb[0])) { stat = iosb[0]; }

	if ($VMS_STATUS_SUCCESS(stat)) { stat = acc.acc$l_finalsts; }

	pthread_cond_destroy(&cndt);
	pthread_mutex_destroy(&mtxt);

	return stat;
}


/*
	Description:

	This routine create a new user in SYSUAF.DAT using a default user.
	The user is added whenever someone tries to logon and the specified
	user is correct but does not exist in SYSUAF.DAT.
	Also, if a command procedure is defined to be run when a user is added
	the return value from that procedure will determine if the user is
	added or not.
	Also an identifier with the name of the username and UIC is created.

	INPUT:		sess handler
			pointer to ldap handler
			RAB
			username to add

	OUTPUT:		SS$_NORMAL is the user was created
			error if something went wrong.
*/

int uaf_add_user(LGI_SESS_T *sess, LDAP **ldap, struct RAB *rab, char *username) {
	int			stat = SS$_NORMAL;
	char			userkey[33] = "DEFAULT";
	char			defdev[33] = ".USERS"; /* save first byte for length */
	char			defdir[65] = ".USER";
	char			owner[33] = ".Unknown LDAP user";
	short unsigned		uic[2] = { 01000, 01000 }; /* default UIC */
	UAF			uafrec;

	stat = get_ldap_user_data(ldap, sess->dn, username, owner + 1, defdev + 1, defdir + 1, uic);

	if (!$VMS_STATUS_SUCCESS(stat)) { return stat; }

	defdev[0] = strlen(defdev + 1);
	defdir[0] = strlen(defdir + 1);
	owner[0] = strlen(owner + 1);

	do_trnlnm("LGI_LDAP_ADD_USER_DEFAULT", userkey, sizeof(userkey) - 1, PSL$C_EXEC);

	rab->rab$l_kbf = (char *)&userkey;
	rab->rab$b_ksz = strlen(userkey);
	rab->rab$b_krf = 0;
	rab->rab$b_rac = RAB$C_KEY;
	rab->rab$l_rop = RAB$M_NLK;
	rab->rab$l_ubf = (char *)&uafrec;
	rab->rab$w_usz = sizeof(uafrec);

	stat = sys$get(rab, 0, 0); /* get the default user */

	if ($VMS_STATUS_SUCCESS(stat)) {
		char		*d = uafrec.uaf$t_username;
		char		*s = username;

		memset(uafrec.uaf$t_username, ' ', sizeof(uafrec.uaf$t_username));
		for(; *s != 0; *d++ = toupper(*s), s++);

		uafrec.uaf$w_mem = uic[0];
		uafrec.uaf$w_grp = uic[1];

		for(d = defdev + 1; *d != 0; *d++ = toupper(*d));
		for(d = defdir + 1; *d != 0; *d++ = toupper(*d));

		lock_cluster(&mtxuai, 0);

		stat = sys$put(rab, 0, 0); /* insert the new user */

		if ($VMS_STATUS_SUCCESS(stat)) { /* we must now update the new user using sys$setuai */
			int unsigned	flags;
			int unsigned	pwddate[2] = { 0, 0 };
			UAFPRIV_T	uafpriv;
			ITEM3		itmlstg[] = { { sizeof(flags),      UAI$_FLAGS,  &flags,  0 },
						      { 0, 0, 0, 0 } };
			ITEM3		itmlsts[] = { { sizeof(defdev) - 1, UAI$_DEFDEV, &defdev, 0 },
						      { sizeof(defdir) - 1, UAI$_DEFDIR, &defdir, 0 },
						      { sizeof(owner) - 1,  UAI$_OWNER,  &owner,  0 },
						      { sizeof(flags),      UAI$_FLAGS,  &flags,  0 },
						      { sizeof(pwddate),    UAI$_PWD_DATE, &pwddate, 0 },
						      { sizeof(uafpriv),    UAI$_USER_DATA, &uafpriv, 0 },
						      { 0, 0, 0, 0 } };
			$DESCRIPTOR(usrnamdsc, username);
			usrnamdsc.dsc$w_length = strlen(username);

			stat = sys$getuai(0, 0, &usrnamdsc, &itmlstg, 0, 0, 0); /* get current flags */

			if ($VMS_STATUS_SUCCESS(stat)) {
				char		actrtn[256];

				if ($VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_ADD_USER_ACTION", actrtn, sizeof(actrtn) - 1, PSL$C_EXEC))) {

					if (!$VMS_STATUS_SUCCESS(stat = do_command(actrtn, 6, username, defdev + 1, defdir + 1, uic))) {

						lib$signal(LLSRV_ADDUSERCMDERR, 2, actrtn, username);

					}
				}

				if ($VMS_STATUS_SUCCESS(stat)) {

					strncpy(uafpriv.dn, sess->dn, sizeof(uafpriv.dn));
					strncpy(uafpriv.uid, username, sizeof(uafpriv.uid));

					do_audit(NSA$C_SYSUAF_ADD, username, NULL, uafpriv.dn);

					flags &= ~UAI$M_DISACNT;     /* remove disuser flag */
					flags &= ~UAI$M_PWD_EXPIRED; /* password not expired */

					uafpriv.version.minver = UAFPRIV_MIN;
					uafpriv.version.majver = UAFPRIV_MAJ;
				       	uafpriv.magic = UAFPRIV_MAGIC;

					stat = sys$setuai(0, 0, &usrnamdsc, &itmlsts, 0, 0, 0);

					if ($VMS_STATUS_SUCCESS(stat)) {

						stat = sys$add_ident(&usrnamdsc, uafrec.uaf$l_uic, 0, 0);

						unlock_cluster(&mtxuai, 0);

						lib$signal(LLSRV_ADDUSER, 2, username, sess->dn, stat);

						return SS$_NORMAL;
					}
				}
                        }

			/*
				If we come here something went wrong. We must delete the user again.
			*/

			rab->rab$l_rop = 0;
			rab->rab$l_kbf = (char *)&uafrec.uaf$t_username;
			rab->rab$b_ksz = strlen(username);

			sys$get(rab, 0, 0);    /* lock the new user so we can delete it */
			sys$delete(rab, 0, 0); /* remove the user again */

		}

		unlock_cluster(&mtxuai, 0);
	}

	lib$signal(LLSRV_ADDUSERERR, 1, username, stat);

	return stat;
}


/*
	Description:

	This routine verifies that a user exists in LDAP.

	INPUT:		pointer to ldap handler
			DN
			uid

	OUTPUT:		LLSRV_LDAPUSEREXIST if the user exist in LDAP
			LLSRV_LDAPUSERMISS if the user is missing in LDAP
*/

int check_user_exists_in_ldap(LDAP **ldap, char *dn, char *uid) {
	int			stat;
	int			retstat = LLSRV_LDAPUSEREXIST;
	int			msgid;
	char			filter[64] = "";
	char			*attrs[] = { "uid", NULL };
	LDAPMessage		*res = NULL;

	if (*ldap == NULL) { return retstat; }

	pthread_mutex_lock(&ldapmtx);

	sprintf(filter, "(uid=%s)", uid);

	if ((msgid = ldap_search(*ldap, dn, LDAP_SCOPE_BASE, filter, attrs, 0)) > 0) {
		struct timeval          to = { 0, 0 };

		while ((stat = ldap_result(*ldap, msgid, 1, &to, &res)) == 0) {
			struct timespec         wait = { 0, 50000 };

			pthread_mutex_unlock(&ldapmtx);
			pthread_delay_np(&wait);
			pthread_mutex_lock(&ldapmtx);

		}

		if (stat > 0) {
			LDAPMessage             *e = (void *)1; /* in case ldap_first_entry goes wrong... */

			if ((e = ldap_first_entry(*ldap, res)) == NULL) {

				retstat = LLSRV_LDAPUSERMISS; /* LDAP communication was ok but the DN is missing */

			}  else {

				ldap_msgfree(e);

			}

			ldap_msgfree(res);

			pthread_mutex_unlock(&ldapmtx);

			return retstat;
		}
	}

	ldap_unbind(*ldap);
	*ldap = NULL;

	pthread_mutex_unlock(&ldapmtx);

	return retstat;
}


typedef struct remw_s {
	pthread_cond_t	*cndtw;
	int		flag;
} REMW_T;


/*
	Description:

	This AST routine is invoked when the remove user thread should run.

	INPUT:		structure used to wakeup the thread

	OUTPUT:		none
*/

void wake_up_rem(REMW_T *remw) {

	remw->flag = 1;
	pthread_cond_signal_int_np(remw->cndtw);

}


/*
	Description:

	This thread is responsible to verify all users in SYSUAF.DAT that
	was created by this program.
	If the DN for each of these users is missing in LDAP the user is	
	removed.
	A command procedure can be defined to run for every users that is 
	to be removed.

	INPUT:		none

	OUTPUT:		none
*/

void *rem_thr(void *s) {
	int			stat;
	LDAP			*ldap = NULL;
	struct FAB		fab = cc$rms_fab;
	struct RAB		rab = cc$rms_rab;
	pthread_cond_t		cndtw = PTHREAD_COND_INITIALIZER;
	pthread_mutex_t		mtxtw = PTHREAD_MUTEX_INITIALIZER;
	REMW_T			remw;

	lib$signal(LLSRV_THRSTARTED, 1, "THREAD_REM_SYSUAF");

	pthread_cond_init(&cndtw, NULL);
	pthread_mutex_init(&mtxtw, NULL);

	remw.cndtw = &cndtw;

	while (1) {
		int unsigned		abstime[2];
		char			runtime[256] = "TOMORROW + 00:15:00";

		do_trnlnm("LGI_LDAP_REMOVE_USER_RUN", runtime, sizeof(runtime) - 1, PSL$C_EXEC);

		stat = parse_run_time(runtime, abstime);

		if (!$VMS_STATUS_SUCCESS(stat)) {

			lib$signal(LLSRV_REMTERM);

			pthread_detach(pthread_self());
			pthread_exit(NULL);

		}

		remw.flag = 0;

		stat = sys$setimr(0, &abstime, &wake_up_rem, &remw, 0);

		while (remw.flag == 0) { pthread_cond_wait(&cndtw, &mtxtw); }

		if (ldap == NULL) {

			if ((ldap = open_ldap()) != NULL) {

				lib$signal(LLSRV_LDAPCONN, 0);

			}
		}

		if (!$VMS_STATUS_SUCCESS(stat = open_sysuaf(&fab, &rab))) {

			lib$signal(LLSRV_ERRACCSYSUAF, 1, "THREAD_REM_SYSUAF", stat);
			pthread_detach(pthread_self());
			pthread_exit(NULL);

		}

		lib$signal(LLSRV_REMUSRRUN, 0);

		do {
			UAFACT_T		*uae;
			UAF			uafrec;

			rab.rab$l_rop = RAB$M_NLK; /* do not lock the record */
			rab.rab$l_ubf = (char *)&uafrec;
			rab.rab$w_usz = sizeof(uafrec);

			stat = sys$get(&rab, 0, 0); /* get the user */

			if (stat == RMS$_NORMAL) {
				int		stat;
				int		uafprivlen;
				short unsigned	uic[2];
				char		defdev[33];
				char		defdir[65];
				UAFPRIV_T	uafpriv;
				ITEM3		itmlst[] = { { sizeof(uafpriv), UAI$_USER_DATA, &uafpriv, &uafprivlen },
		      					     { sizeof(defdev) - 1, UAI$_DEFDEV, &defdev, 0 },
							     { sizeof(defdir) - 1, UAI$_DEFDIR, &defdir, 0 },
							     { sizeof(uic), UAI$_UIC, &uic, 0 },
							     { 0, 0, 0, 0 } };
				$DESCRIPTOR(usrnamdsc, uafrec.uaf$t_username);

				stat = sys$getuai(0, 0, &usrnamdsc, &itmlst, 0, 0, 0);

				if ($VMS_STATUS_SUCCESS(stat) && uafprivlen > 0 && uafpriv.magic == UAFPRIV_MAGIC) {

					stat = check_user_exists_in_ldap(&ldap, uafpriv.dn, uafpriv.uid);

					if (stat == LLSRV_LDAPUSERMISS) { /* user does not exist in LDAP anymore with the DN */

						rab.rab$l_rop = 0; /* lock the record this time */
						rab.rab$b_rac = RAB$C_KEY;
						rab.rab$l_kbf = (char *)&uafrec.uaf$t_username;
						rab.rab$b_ksz = sizeof(uafrec.uaf$t_username);

						stat = sys$get(&rab, 0, 0); /* get the user and lock it */

						if (stat == RMS$_NORMAL) {
							char		actrtn[256];
                                                        char		username[sizeof(uafrec.uaf$t_username) + 1];

							if ($VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_REMOVE_USER_ACTION", actrtn,
											  sizeof(actrtn) - 1, PSL$C_EXEC))) {

								defdev[defdev[0] + 1] = 0;
								defdir[defdir[0] + 1] = 0;

			       	      				stat = do_command(actrtn, 30, uafrec.uaf$t_username, defdev + 1, defdir + 1, uic);

							}

							sys$rem_ident((uic[1] << 16) | uic[0]);

							stat = sys$delete(&rab, 0, 0);

							strncpy(username, uafrec.uaf$t_username, sizeof(uafrec.uaf$t_username) - 1);
							username[sizeof(uafrec.uaf$t_username)] = 0;

							do_audit(NSA$C_SYSUAF_DELETE, username, uafpriv.dn, NULL);
						}

						rab.rab$b_rac = RAB$C_SEQ;
					}

					stat = RMS$_NORMAL;
				}
			}

		} while (stat == RMS$_NORMAL);

		lib$signal(LLSRV_REMUSRDONE, 0);

		close_sysuaf(&fab, &rab);
	}

	return NULL;
}


/*
	Description:

	This thread is responsible to add users to SYSUAF.DAT.
	
	INPUT:		none

	OUTPUT:		none
*/

void *add_thr(void *s) {
	int		stat;
	char		ldaphost[256];
	char		basedn[256];
	LDAP		*ldap = NULL;
	struct FAB	fab = cc$rms_fab;
	struct RAB	rab = cc$rms_rab;

	lib$signal(LLSRV_THRSTARTED, 1, "THREAD_ADD_SYSUAF");

	if (!$VMS_STATUS_SUCCESS(stat = open_sysuaf(&fab, &rab))) {

		add_user_ok = 0;

		lib$signal(LLSRV_ERRACCSYSUAF, 1, "THREAD_ADD_SYSUAF", stat);

		pthread_detach(pthread_self());
		pthread_exit(NULL);

	}

	while (1) {
		UAFACT_T		*uae;

		pthread_cond_wait(&cndaddw, &mtxaddw);

		if (ldap == NULL) {

			if ((ldap = open_ldap()) != NULL) {

				lib$signal(LLSRV_LDAPCONN, 0);

			}
		}

		while ((uae = remove_qh(&addqh)) != NULL) {

			uae->stat = uaf_add_user(uae->sess, &ldap, &rab, uae->username);

			uae->sess->flags.condok = 1;
			pthread_cond_signal_int_np(&uae->sess->cndw);

		}

	}

	return NULL;
}


/*
	Description:

	This thread is responsible to handle identity verification and
	password sync.

	INPUT:		none

	OUTPUT:		none
*/

void *mod_thr(void *s) {

	lib$signal(LLSRV_THRSTARTED, 1, "THREAD_MOD_SYSUAF");

	while (1) {
		UAFACT_T		*uae;

		pthread_cond_wait(&cndw, &mtxw);

		while ((uae = remove_qh(&verqh)) != NULL) {

			uae->stat = uaf_verify(uae->sess, uae->username);
			uae->sess->flags.condok = 1;
			pthread_cond_signal_int_np(&uae->sess->cndw);

		}

		while ((uae = remove_qh(&pwdqh)) != NULL) {

			uae->stat = uaf_pwd_synch(uae->sess, uae->password);
			uae->sess->flags.condok = 1;
			pthread_cond_signal_int_np(&uae->sess->cndw);

		}
	}

	return NULL;
}


/*
	Description:

	This routine is used to execute a User Action Entry.
	The calling thread will continue when the action has been completed.

	INPUT:		sess handler
			User Action Entry
			queue header

	OUTPUT:		status from action routine.
*/

int do_uae(LGI_SESS_T *sess, UAFACT_T *uae, void *qh) {

	uae->sess = sess;

	insert_qt(qh, uae);

	sess->flags.condok = 0;
	pthread_cond_signal_int_np(&cndw);
	while (!sess->flags.condok) { pthread_cond_wait(&sess->cndw, &sess->mtxw); }

	return uae->stat;
}


/*
	Description:

	This routine is used to synchronize the password  on the host with the password
	on LDAP server.

	INPUT:		sess handler
			credential (password)

	OUTPUT:		status from action routine
*/

int synch_password(LGI_SESS_T *sess, char *credential) {
	int			stat;
        UAFACT_T		*uae = alloc_qe(sizeof(UAFACT_T));

	strncpy(uae->password, credential, sizeof(uae->password) - 1);

	stat = do_uae(sess, uae, &pwdqh);

	free_qe(sizeof(UAFACT_T), uae);

	return stat;

}


/*
	Description:

	This routine is used to verify the identity (username) on the host.

	INPUT:		sess handler
			credential (password)

	OUTPUT:		status from action routine
*/

int verify_username(LGI_SESS_T *sess, char *identity, int create) {
	int			stat;
	char			dummy[256];
        UAFACT_T		*uae = alloc_qe(sizeof(UAFACT_T));

	strncpy(uae->username, identity, sizeof(uae->username) - 1);

	stat = do_uae(sess, uae, &verqh);

	if (create && add_user_ok && !$VMS_STATUS_SUCCESS(stat)) {

		insert_qt(&addqh, uae);

		sess->flags.condok = 0;
		pthread_cond_signal_int_np(&cndaddw);
		while (!sess->flags.condok) { pthread_cond_wait(&sess->cndw, &sess->mtxw); }

		stat = do_uae(sess, uae, &verqh);

	}

	free_qe(sizeof(UAFACT_T), uae);

	return stat;
}


/*
	Description:

	This routine is inoked when the program is started.

	INPUT:		none

	OUTPUT:		none
*/

int uaf_init() {
	char			dummy[256];

	pthread_mutex_init(&mtxw, NULL);
	pthread_cond_init(&cndw, NULL);

	pthread_create(&thrmod, NULL, &mod_thr, NULL);

	if ($VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_ADD_USER", dummy, sizeof(dummy) - 1, PSL$C_EXEC))) {

		pthread_mutex_init(&mtxaddw, NULL);
		pthread_cond_init(&cndaddw, NULL);

		if (pthread_create(&thradd, NULL, &add_thr, NULL) == 0) { add_user_ok = 1; }

	}

	if ($VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_REMOVE_USER", dummy, sizeof(dummy) - 1, PSL$C_EXEC))) {

		pthread_create(&thrrem, NULL, &rem_thr, NULL);

	}

	return SS$_NORMAL;
}
